/*
 *      Copyright (C) 2009-2010 Antti Ajanki <antti.ajanki@iki.fi>
 *
 *  This Program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This Program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RTMPDump; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *  http://www.gnu.org/copyleft/gpl.html
 *
 */

/*
 * This file implements the functionality from the YLE Areena media
 * player (http://areena-beta.yle.fi/player/Application.swf) that is
 * necessary for downloading streams.
 *
 * A "normal" RTMP media player extract the name of the stream
 * (playpath) from the RTMP URL, and invokes the remote procedure
 * play() on it after connecting to the server. YLE Areena has two
 * features that make downloading more difficult. When the player
 * connects to the server, the server sends a challenge and the player
 * has to respond correctly, otherwise the server refuses to talk to
 * the player. Secondly, the playpath is not available on the video
 * web page. Instead, the player has to send clipID, the final
 * component of the web page URL, to the server and extract the
 * playpath from the server's response.
 *
 * The initial communication after the handshake looks like this:
 *
 * client -> server: invoke (RTMP packet 0x14): connect()
 *
 * S -> C: invoke: authenticationDetails(..., randomAuth: 12345)
 *
 * C -> S: flex message (0x11): authenticateRandomNumber((randomAuth + 447537687) % 6834253)
 *
 * S -> C: invoke: randomNumberAuthenticated()
 *
 * C -> S: flex message: requestData("e0", "session/authenticate/1")
 *
 * C -> S: flex message: requestData("e1", "clips/info/999") (= clip ID from the URL)
 *           (or requestData("e1", "streams/info/999/lang/fi") for live streams)
 *           (or requestData("e1", "channels/info/999/lang/fi") for net radios)
 *
 * S -> C: invoke: rpcResult(..., "status": "200", ..., "url": "rtmp://host/playpath")
 *
 * C -> S: flex message: requestData("e2", "clips/related/999")
 *           (or requestData("e2", "channels/lang/fi/geo/FI") for net radios)
 *
 * C -> S: flex message: requestData("e3", "clips/featured/999")
 *           (or requestData("e3", "categories/lang/fi") for net radios)
 *
 * The e0, e2, e3 steps are not strictly required to play the stream
 * (and their results can be ignored), but they are related to some
 * auxiliary features in the Areena Flash player.
 *
 * After getting the playpath in the response to requestData("e1"),
 * the player can continue downloading the stream normally.
 *
 */

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <langinfo.h>

#include "json-c-0.9/json.h"
#include "json-c-0.9/json_object_private.h"

#include "librtmp/log.h"
#include "librtmp/http.h"
#include "yle.h"

#ifndef WIN32
#include <iconv.h>
#endif

#define DEFAULTDESTFILE "YLEAreena"

#define FREE_AND_NULL(x) { if(x) { free(x); x = NULL; } }
#define STR2AVAL(av,str) { av.av_val = str; av.av_len = strlen(av.av_val); }
#define SAVC(x) static const AVal av_##x = AVC(#x)

SAVC(_result);
SAVC(connect);
SAVC(authenticationDetails);
SAVC(randomNumberAuthenticated);
SAVC(rpcResult);
SAVC(rpcError);
SAVC(locatedInBroadcastTerritory);
SAVC(randomAuth);
SAVC(tvFeeActivated);
SAVC(authenticateRandomNumber);
SAVC(e0);
SAVC(e1);
SAVC(e2);
SAVC(e3);
SAVC(requestData);

typedef struct MemoryBuffer
{
  char *ptr;
  size_t len;
  size_t allocated;
} MemoryBuffer;

static void ConstructYLEPlaypath(const AVal *app, const AVal *playpath, AVal *out);
static int ExecuteAuthenticationDetails(RTMP *r, AMFObject *proplist);
static int ExecuteRandomNumberAuthenticated(RTMP *r);
static int ExecuteRPCResult(RTMP *r, AMFObject *obj);
static int SessionAuthenticate(RTMP *r);
static int RequestData(RTMP *r, const AVal *request, const AVal *id);
static int ClipsInfo(RTMP *r);
static int RelatedClips(RTMP *r);
static int FeaturedClips(RTMP *r);
static int ChannelsList(RTMP *r);
static int CategoriesList(RTMP *r);
static char *SaneFileName(const char *filename, int vfatchars);
static char *EncodeFromUTF8(char *string, const char *to);
static char *FormatYLEDateTime(char *datestring);
static int DateHasTime(const char *datestring);
static size_t CollectStream(void *ptr, size_t size, size_t nmemb, void *stream);
static char *GetXMLNodeContent(const char *xmldoc, const char *node);
static int LoadAkamaiStreamParameters(RTMP *r, const char *xmlurl);
static void ConstructAkamaiParams(const char *streamname,
                                  const char *authparams,
                                  const char *appname,
                                  const char *hostname,
                                  const char *hostip,
                                  AVal *playpath,
                                  AVal *app,
                                  AVal *tcUrl);

// Parse Areena Player parameter string. Returns a new YLEStream
// which the caller must free, or NULL on error.
YLEStream *YLE_ParseAreenaParams(const char *areenaParams) {
  RTMP_Log(RTMP_LOGDEBUG, "params: %s", areenaParams);

  if (!areenaParams)
    return NULL;

  YLEStream *stream = malloc(sizeof(YLEStream));
  if (!stream)
    return NULL;

  stream->clipID = NULL;
  stream->liveStreamID = NULL;
  stream->radioID = NULL;
  stream->language = NULL;
  stream->title = NULL;
  stream->tvFeeRequired = FALSE;
  stream->tvFeeActivated = FALSE;
  stream->locatedInBroadcastTerritory = FALSE;
  stream->isAkamaiStream = FALSE;
  stream->redirected = FALSE;

  char *vars = strdup(areenaParams);
  char *token = strtok(vars, ";");

  while (token) {
    RTMP_Log(RTMP_LOGDEBUG, "Areena Player param: %s", token);

    if (strncmp(token, "clip:", 5) == 0) {
      stream->clipID = strdup(&token[5]);
    } else if (strncmp(token, "stream:", 7) == 0) {
      stream->liveStreamID = strdup(&token[7]);
    } else if (strncmp(token, "channel:", 8) == 0) {
      stream->radioID = strdup(&token[8]);
    } else if (strncmp(token, "language:", 9) == 0) {
      stream->language = strdup(&token[9]);
    }

    token = strtok(NULL, ";");
  }

  free(vars);

  return stream;
}

// Returns -1 for unhandled, 0 for OK/Failed/error, 1 for 'Stop or Complete'
int YLE_ExecuteInvokedMethod(RTMP *r, AMFObject *obj) {
  AVal method;
  AMFProp_GetString(AMF_GetProp(obj, NULL, 0), &method);

  if (AVMATCH(&method, &av__result)) {
    AVal methodInvoked = r->m_methodCalls[0].name;

    if (AVMATCH(&methodInvoked, &av_connect)) {
      RTMP_DropRequest(r, 0, FALSE);

      RTMP_Log(RTMP_LOGDEBUG, "%s, received result for method call <%s>",
          __FUNCTION__, methodInvoked.av_val);

      // Like connection handling in HandleInvoke() but don't send
      // CreateStream.

      RTMP_SendServerBW(r);
      //RTMP_SendCtrl(r, 3, 0, 300);

      free(methodInvoked.av_val);

      return 0;
    }

  } else if (AVMATCH(&method, &av_authenticationDetails)) {
    AMFObject list;
    AMFProp_GetObject(AMF_GetProp(obj, NULL, 3), &list);
    ExecuteAuthenticationDetails(r, &list);
    return 0;

  } else if (AVMATCH(&method, &av_randomNumberAuthenticated)) {
    ExecuteRandomNumberAuthenticated(r);
    return 0;

  } else if (AVMATCH(&method, &av_rpcResult)) {
    if (!ExecuteRPCResult(r, obj) || r->m_yleStream->redirected)
      return 1;
    else
      return 0;

  } else if (AVMATCH(&method, &av_rpcError)) {
    RTMP_Log(RTMP_LOGERROR, "RTMP server returned RPC error");
    return 0;
  }

  return -1;
}

// Takes ownership of areena pointer.
void YLE_SetupYLEStream(RTMP *r, YLEStream *yleStream) {
  // If more than one of clip, live stream or radio IDs are
  // specified, prefer live stream to clip, and clip to radio
  if (yleStream->liveStreamID && yleStream->clipID) {
    RTMP_Log(RTMP_LOGDEBUG, "Live stream overrides clip.");
    FREE_AND_NULL(yleStream->clipID);
  }
  if (yleStream->clipID && yleStream->radioID) {
    RTMP_Log(RTMP_LOGDEBUG, "Clip overrides radio.");
    FREE_AND_NULL(yleStream->radioID);
  }
  if (yleStream->liveStreamID && yleStream->radioID) {
    RTMP_Log(RTMP_LOGDEBUG, "Live stream overrides radio.");
    FREE_AND_NULL(yleStream->radioID);
  }

  if (yleStream->liveStreamID)
    RTMP_Log(RTMP_LOGDEBUG, "LivesID  : %s", yleStream->liveStreamID);
  if (yleStream->clipID)
    RTMP_Log(RTMP_LOGDEBUG, "ClipID   : %s", yleStream->clipID);
  if (yleStream->radioID)
    RTMP_Log(RTMP_LOGDEBUG, "RadioID  : %s", yleStream->radioID);

  r->m_yleStream = yleStream;
  r->invokeHook = YLE_ExecuteInvokedMethod;
}

int ExecuteAuthenticationDetails(RTMP *r, AMFObject *proplist) {
  long authResult = -1;
  int i;
  
  for (i=0; i<AMF_CountProp(proplist); i++) {
    AVal name;
    AMFObjectProperty *prop = AMF_GetProp(proplist, NULL, i);
    AMFProp_GetName(prop, &name);

    if (AVMATCH(&name, &av_locatedInBroadcastTerritory)) {
      r->m_yleStream->locatedInBroadcastTerritory = AMFProp_GetBoolean(prop);
    } else if (AVMATCH(&name, &av_randomAuth)) {
      authResult = ((long)AMFProp_GetNumber(prop) + 447537687) % 6834253;
    } else if (AVMATCH(&name, &av_tvFeeActivated)) {
      r->m_yleStream->tvFeeActivated = AMFProp_GetBoolean(prop);
    }
  }

  if (authResult != -1) {
    RTMPPacket packet;
    char pbuf[128], *pend = pbuf+sizeof(pbuf);

    packet.m_nChannel = 0x03;   // control channel
    packet.m_headerType = RTMP_PACKET_SIZE_LARGE;
    packet.m_packetType = 0x11; // FLEX MESSAGE
    packet.m_nTimeStamp = RTMP_GetTime();
    packet.m_nInfoField2 = 0;
    packet.m_hasAbsTimestamp = 0;

    packet.m_body = pbuf + RTMP_MAX_HEADER_SIZE;
    char *enc = packet.m_body;
    *enc++ = 0x00;   // Unknown
    enc = AMF_EncodeString(enc, pend, &av_authenticateRandomNumber);
    enc = AMF_EncodeNumber(enc, pend, 0);
    *enc++ = AMF_NULL;
    enc = AMF_EncodeNumber(enc, pend, (double)authResult);

    packet.m_nBodySize = enc-packet.m_body;

    return RTMP_SendPacket(r, &packet, FALSE);
  }

  return FALSE;
}

int ExecuteRandomNumberAuthenticated(RTMP *r) {
  // It seems to be possible to skip SessionAuthenticate() and do
  // ClipsInfo() directly, but let's follow the Flash player as
  // closely as possible.

  return SessionAuthenticate(r);
  //return ClipsInfo(r);
}

/*
 * Returns concatenation of appinstance (last part of app) and
 * playpath in out.
 *
 * Areena server requires that the appinstance is included into the
 * playpath. I think that this is a bug in the Areena software...
 */
void ConstructYLEPlaypath(const AVal *app, const AVal *playpath, AVal *out) {
  char *slash = memchr(app->av_val, '/', app->av_len);
  if (!slash) {
    out->av_len = playpath->av_len;
    out->av_val = malloc(out->av_len*sizeof(char));
    strncpy(out->av_val, playpath->av_val, out->av_len);
    return;
  }

  AVal appinstance = {slash+1, app->av_len-(slash+1-app->av_val)};
  AVal streamtype = {NULL, 0};
  AVal streamname = {NULL, 0};
  char *colon = memchr(playpath->av_val, ':', playpath->av_len);
  if (colon) {
    streamtype.av_val = playpath->av_val;
    streamtype.av_len = colon - playpath->av_val + 1;

    streamname.av_val = colon+1;
    streamname.av_len = playpath->av_len - (colon-playpath->av_val) - 1;
  } else {
    streamtype.av_val = "";
    streamtype.av_len = 0;

    streamname.av_val = playpath->av_val;
    streamname.av_len = playpath->av_len;
  }
  
  out->av_len = streamtype.av_len + appinstance.av_len + playpath->av_len + 2;
  out->av_val = malloc(out->av_len*sizeof(char));
  snprintf(out->av_val, out->av_len, "%.*s%.*s/%.*s",
           streamtype.av_len, streamtype.av_val,
           appinstance.av_len, appinstance.av_val,
           streamname.av_len, streamname.av_val);
}

int ExecuteRPCResult(RTMP *r, AMFObject *obj) {
  int res = FALSE;
  int jsonstatus = -1;
  AVal callID, jsondata;

  AMFProp_GetString(AMF_GetProp(obj, NULL, 3), &callID);
  AMFProp_GetString(AMF_GetProp(obj, NULL, 4), &jsondata);

  char *message = malloc((jsondata.av_len+1)*sizeof(char));
  memcpy(message, jsondata.av_val, jsondata.av_len);
  message[jsondata.av_len] = '\0';
  struct json_object *json = json_tokener_parse(message);
  free(message);

  if (!is_error(json)) {
    struct json_object_iter outeriter;
    json_object_object_foreachC(json, outeriter) {
      if (!strcmp(outeriter.key, "status")) {
        struct json_object_iter inneriter;
        json_object_object_foreachC(outeriter.val, inneriter) {
          if (!strcmp(inneriter.key, "code")) {
            jsonstatus = json_object_get_int(inneriter.val);
          }
        }
      }
    }

    if (jsonstatus != 200) {
      RTMP_Log(RTMP_LOGWARNING, "Unexpected RPC status code: %d. Trying to continue anyway.", jsonstatus);
    }
  } else if (jsondata.av_len != 0) {
    RTMP_Log(RTMP_LOGWARNING, "Failed to parse JSON string: %*s", jsondata.av_len, jsondata.av_val);
  }

  if (AVMATCH(&callID, &av_e0)) {
    // Response ignored
    res = ClipsInfo(r);

  } else if ((AVMATCH(&callID, &av_e1) && !is_error(json))) { // clips/info/12345
    char *title = NULL;
    char *date = NULL;
    const char *playurl = NULL;
    
    struct json_object_iter outeriter;
    json_object_object_foreachC(json, outeriter) {
      if (!strcmp(outeriter.key, "data")) {
        struct json_object_iter inneriter;
        json_object_object_foreachC(outeriter.val, inneriter) {
          if (!strcmp(inneriter.key, "url")) {
            playurl = json_object_get_string(inneriter.val);
            if (playurl) {
              AVal parsedHost = {NULL, 0};
              unsigned int parsedPort = 0;
              AVal parsedPlaypath = {NULL, 0};
              AVal parsedApp = {NULL, 0};
              int parsedProtocol = RTMP_PROTOCOL_UNDEFINED;

              RTMP_Log(RTMP_LOGDEBUG, "Got the real stream URL: %s", playurl);
              
              if (r->m_yleStream->isAkamaiStream) {
                RTMP_Log(RTMP_LOGDEBUG, "Akamai playpath is parsed later...");
                res = TRUE;
              } else if (!RTMP_ParseURL(playurl, &parsedProtocol, &parsedHost,
                                        &parsedPort, &parsedPlaypath, &parsedApp)) {
                RTMP_Log(RTMP_LOGERROR, "Couldn't parse stream url %s!", playurl);
              } else {
                // FIXME: old r->Link.playpath may be leaked 
                ConstructYLEPlaypath(&parsedApp, &parsedPlaypath, &r->Link.playpath);
                FREE_AND_NULL(parsedPlaypath.av_val);
                parsedPlaypath.av_len = 0;

                RTMP_Log(RTMP_LOGDEBUG, "New playpath   : %.*s",
                         r->Link.playpath.av_len, r->Link.playpath.av_val);

                if (!AVMATCH(&parsedHost, &r->Link.hostname)) {
                  RTMP_Log(RTMP_LOGDEBUG, "Redirected to another server: %.*s",
                           parsedHost.av_len, parsedHost.av_val);

                  // FIXME: old value may be leaked
                  r->Link.hostname.av_val = malloc(parsedHost.av_len*sizeof(char));
                  r->Link.hostname.av_len = parsedHost.av_len;
                  memcpy(r->Link.hostname.av_val, parsedHost.av_val, parsedHost.av_len);

                  r->m_yleStream->redirected = TRUE;
                }

                res = TRUE;
              }
            }

          } else if (!strcmp(inneriter.key, "seriesTitle")) {
            const char *jsontitle = json_object_get_string(inneriter.val);
            if (jsontitle && *jsontitle) {
              // seriesTitle is preferred to title
              RTMP_Log(RTMP_LOGDEBUG, "Got seriesTitle: %s", jsontitle);
              if (title)
                free(title);
              title = strdup(jsontitle);
            }

          } else if (!strcmp(inneriter.key, "title")) {
            const char *jsontitle = json_object_get_string(inneriter.val);
            if (jsontitle && *jsontitle) {
              RTMP_Log(RTMP_LOGDEBUG, "Got title: %s", jsontitle);
              if (!title)
                title = strdup(jsontitle);
            }

          } else if (!strcmp(inneriter.key, "broadcastDateTime")) {
            const char *jsondate = json_object_get_string(inneriter.val);
            if (jsondate && *jsondate && (strcmp(jsondate, "0") != 0)) {
              // broadcastDateTime is preferred to publicationDateTime
              // expect if broadcastDateTime has no time
              RTMP_Log(RTMP_LOGDEBUG, "Got broadcastDateTime: %s", jsondate);
              if (!date || DateHasTime(jsondate)) {
                if (date)
                  free(date);
                date = strdup(jsondate);
              }
            }

          } else if (!strcmp(inneriter.key, "publicationDateTime")) {
            const char *jsondate = json_object_get_string(inneriter.val);
            if (jsondate && *jsondate && (strcmp(jsondate, "0") != 0)) {
              RTMP_Log(RTMP_LOGDEBUG, "Got publicationDateTime: %s", jsondate);
              if (!date || (!DateHasTime(date) && DateHasTime(jsondate))) {
                if (date)
                  free(date);
                date = strdup(jsondate);
              }
            }

          } else if (!strcmp(inneriter.key, "akamai")) {
            r->m_yleStream->isAkamaiStream = json_object_get_boolean(inneriter.val);
            if (r->m_yleStream->isAkamaiStream) {
              RTMP_Log(RTMP_LOGDEBUG, "This is an Akamai stream");
              r->m_yleStream->redirected = TRUE;
            }

          } else if (!strcmp(inneriter.key, "tvFeeRequired")) {
            r->m_yleStream->tvFeeRequired = json_object_get_boolean(inneriter.val);
          }
        }
      }
    }

    if (res) {
      FREE_AND_NULL(r->m_yleStream->title);

      int datelen = 0;
      if (date) {
        date = FormatYLEDateTime(date);
        datelen = strlen(date)+1;
      }

      if (title) {
        r->m_yleStream->title = malloc((strlen(title)+datelen+1)*sizeof(char));
        strcpy(r->m_yleStream->title, title);
      } else {
        r->m_yleStream->title = malloc((32+datelen+1)*sizeof(char));
        if (r->m_yleStream->liveStreamID) {
          sprintf(r->m_yleStream->title, DEFAULTDESTFILE "-live%s", 
                  r->m_yleStream->liveStreamID);
        } else if (r->m_yleStream->clipID) {
          sprintf(r->m_yleStream->title, DEFAULTDESTFILE "-clip%s",
                  r->m_yleStream->clipID);
        } else if (r->m_yleStream->radioID) {
          sprintf(r->m_yleStream->title, DEFAULTDESTFILE "radio%s",
                  r->m_yleStream->radioID);
        }
      }

      if (date && *date) {
        strcat(r->m_yleStream->title, "-");
        strcat(r->m_yleStream->title, date);
      }

      if (r->m_yleStream->redirected) {
        if (r->m_yleStream->isAkamaiStream && !LoadAkamaiStreamParameters(r, playurl)) {
          RTMP_Log(RTMP_LOGERROR, "Failed to load Akamai parameters");
          r->m_yleStream->redirected = FALSE;
        }

        RTMP_Close(r); // Going to reconnect to another server
        res = FALSE;
      } else {
        if (r->m_yleStream->clipID)
          //20100728: removed RelatedClips(r), response is 404 
          res = FeaturedClips(r);
        else if (r->m_yleStream->radioID)
          res = ChannelsList(r) && CategoriesList(r);
        else
          res = RTMP_SendCreateStream(r);
      }
    }

    FREE_AND_NULL(title);
    FREE_AND_NULL(date);

  } else if (AVMATCH(&callID, &av_e2)) {
    // Response ignored
    res = TRUE;

  } else if (AVMATCH(&callID, &av_e3)) {
    // Response ignored
    res = RTMP_SendCreateStream(r);

  }

  if (json && !is_error(json))
    json_object_put(json);

  return res;
}

int RequestData(RTMP *r, const AVal *request, const AVal *id) {
  RTMPPacket packet;
  char pbuf[256], *pend = pbuf+sizeof(pbuf);

  packet.m_nChannel = 0x03; // control channel
  packet.m_headerType = RTMP_PACKET_SIZE_LARGE;
  packet.m_packetType = 0x11; // FLEX MESSAGE
  packet.m_nTimeStamp = RTMP_GetTime();
  packet.m_nInfoField2 = 0;
  packet.m_hasAbsTimestamp = 0;
  packet.m_body = pbuf + RTMP_MAX_HEADER_SIZE;
  
  char *enc = packet.m_body;
  *enc++ = 0x00;   // Unknown
  enc = AMF_EncodeString(enc, pend, &av_requestData);
  enc = AMF_EncodeNumber(enc, pend, 0.0);
  *enc++ = AMF_NULL;
  enc = AMF_EncodeString(enc, pend, id);
  enc = AMF_EncodeString(enc, pend, request);

  if (!enc) {
    RTMP_Log(RTMP_LOGERROR, "Buffer too short in %s", __FUNCTION__);
    return FALSE;
  }

  packet.m_nBodySize = enc-packet.m_body;
  
  RTMP_Log(RTMP_LOGDEBUG, "Sending requestData (%*s) %*s",
      id->av_len, id->av_val, request->av_len, request->av_val);

  return RTMP_SendPacket(r, &packet, FALSE);
}

int SessionAuthenticate(RTMP *r) {
  char *authcmd, *authval;
  AVal auth;

  if (r->Link.auth.av_len > 0)
    authval = r->Link.auth.av_val;
  else
    authval = "1";

  int buflen = 22 + strlen(authval);
  authcmd = malloc(buflen*sizeof(char));
  snprintf(authcmd, buflen, "session/authenticate/%s", authval);
  STR2AVAL(auth, authcmd);

  int res = RequestData(r, &auth, &av_e0);

  free(authcmd);

  return res;
}

int ClipsInfo(RTMP *r) {
  const int buflen = 64;
  char buf[buflen];
  int written;

  if (r->m_yleStream->liveStreamID) {
    written = snprintf(buf, buflen, "streams/info/%s/lang/%s",
                       r->m_yleStream->liveStreamID, 
                       r->m_yleStream->language ? r->m_yleStream->language : "");
  } else if (r->m_yleStream->clipID) {
    written = snprintf(buf, buflen, "clips/info/%s", r->m_yleStream->clipID);
  } else if (r->m_yleStream->radioID) {
    written = snprintf(buf, buflen, "channels/info/%s/lang/%s",
                       r->m_yleStream->radioID,
                       r->m_yleStream->language ? r->m_yleStream->language : "");
  } else {
    return FALSE;
  }

  if (written >= buflen) {
    RTMP_Log(RTMP_LOGERROR, "Buffer too short in %s", __FUNCTION__);
    return FALSE;
  }
    
  AVal cmd;
  STR2AVAL(cmd, buf);

  return RequestData(r, &cmd, &av_e1);
}

int RelatedClips(RTMP *r) {
  if (!r->m_yleStream->clipID)
    return FALSE;
  
  const int buflen = 64;
  char buf[buflen];

  if (snprintf(buf, buflen, "clips/related/%s", r->m_yleStream->clipID) >= buflen) {
    RTMP_Log(RTMP_LOGERROR, "Buffer too short in %s", __FUNCTION__);
    return FALSE;
  }

  AVal cmd;
  STR2AVAL(cmd, buf);

  return RequestData(r, &cmd, &av_e2);
}

int FeaturedClips(RTMP *r) {
  if (!r->m_yleStream->clipID)
    return FALSE;

  const int buflen = 64;
  char buf[buflen];

  if (snprintf(buf, buflen, "clips/featured/%s", r->m_yleStream->clipID) >= buflen) {
    RTMP_Log(RTMP_LOGERROR, "Buffer too short in %s", __FUNCTION__);
    return FALSE;
  }

  AVal cmd;
  STR2AVAL(cmd, buf);

  return RequestData(r, &cmd, &av_e3);
}

int ChannelsList(RTMP *r) {
  const char *lang = r->m_yleStream->language;
  if (!lang)
    lang = "fi";

  const int buflen = 64;
  char buf[buflen];

  if (snprintf(buf, buflen, "channels/lang/%s/geo/FI", lang) >= buflen) {
    RTMP_Log(RTMP_LOGERROR, "Buffer too short in %s", __FUNCTION__);
    return FALSE;
  }

  AVal cmd;
  STR2AVAL(cmd, buf);
  
  return RequestData(r, &cmd, &av_e2);
}

int CategoriesList(RTMP *r) {
  const char *lang = r->m_yleStream->language;
  if (!lang)
    lang = "fi";

  const int buflen = 64;
  char buf[buflen];

  if (snprintf(buf, buflen, "categories/lang/%s", lang) >= buflen) {
    RTMP_Log(RTMP_LOGERROR, "Buffer too short in %s", __FUNCTION__);
    return FALSE;
  }

  AVal cmd;
  STR2AVAL(cmd, buf);

  return RequestData(r, &cmd, &av_e3);
}

char *YLE_GetSuggestedFileName(RTMP *r, int vfatchars) {
  char *sanename = SaneFileName(r->m_yleStream->title, vfatchars);

  // Areena streams are always flv files regardless of the extension
  // in the URL.
  char *extension = ".flv";

  int buflen = strlen(sanename) + strlen(extension) + 1;
  char *finalname = malloc(buflen*sizeof(char));
  strcpy(finalname, sanename);
  strcat(finalname, extension);

  free(sanename);

  return finalname;
}

char *EncodeFromUTF8(char *string, const char *tocode) {
#ifdef WIN32

  return strdup(string);

#else

  size_t inlen = strlen(string);
  size_t outlen = 2*inlen;
  char codetranslit[32];

  snprintf(codetranslit, 32, "%s//TRANSLIT", tocode);
  codetranslit[31] = '\0';

  iconv_t cd = iconv_open(codetranslit, "UTF-8");
  if (cd == (iconv_t)(-1)) {
    RTMP_Log(RTMP_LOGERROR, "Encoding failed: %m");
    return NULL;
  }

  RTMP_Log(RTMP_LOGDEBUG, "Encoding %s from UTF8 to %s", string, tocode);

  char *encoded = (char *)malloc(outlen*sizeof(char));
  if (!encoded) {
    iconv_close(cd);
    return NULL;
  }

  char *out = encoded;
  if (iconv(cd, &string, &inlen, &out, &outlen) == (size_t)(-1)) {
    RTMP_Log(RTMP_LOGWARNING, "Encoding failed: %m");
  }

  iconv_close(cd);

  if (outlen < 1)
    out[-1] = '\0';
  else
    out[0] = '\0';
  outlen--;
  
  return encoded;

#endif
}

char *SaneFileName(const char *filename, int vfatchars) {
  const char *start, *end;
  char *newname;

  if (!filename)
    return strdup(DEFAULTDESTFILE);

  // Remove dots from the beginning.
  start = filename;
  while (*start == '.')
    start++;

  // Windows does not accept names that end in a dot.
  end = filename+strlen(filename);
  while (end >= start && *end == '.')
    end--;

  int newlen = end-start;
  newname = malloc((newlen+1)*sizeof(char));
  if (!newname)
    return NULL;

  strncpy(newname, start, newlen);
  newname[newlen] = '\0';

  char *encoded = EncodeFromUTF8(newname, nl_langinfo(CODESET));

  if (encoded) {
    free(newname);
    newname = encoded;
  }

  // The following characters are illegal ('/' on Unix filesystems and
  // the rest on NTFS and on VFAT) or may cause problems. Replace them
  // with underscores.
  const char *excludechars;
  if (vfatchars)
    excludechars = "\"*/:<>?|";
  else
    excludechars = "*/|";

  char *p = strpbrk(newname, excludechars);
  while (p) {
    *p = '_';
    p = strpbrk(p+1, excludechars);
  }

  if (*newname == '\0') {
    free(newname);
    newname = strdup(DEFAULTDESTFILE);
  }

  return newname;
}

// Open origname for writing if it does not exists. If it exists try
// alternative names by appending -X, where X is an integer, to the
// original name. If a file with an alternative name is opened, its
// name is put in *newname, otherwise *newname is NULL.
FILE *OpenAvailableName(const char *origname, char **newname) {
  *newname = NULL;
#ifdef WIN32
  int fd  = open(origname, O_CREAT | O_EXCL | O_RDWR | O_BINARY);
#else
  int fd  = open(origname, O_CREAT | O_EXCL | O_RDWR,
                 S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif
  if (fd != -1) {
    return fdopen(fd, "r+b");
  }

  RTMP_Log(RTMP_LOGDEBUG, "File %s already exists. Trying alternative file names.", origname);

  int lasterror = errno;
  int i = 1;
  const char *ext = strrchr(origname, '.');
  if (!ext)
    ext = strchr(origname, '\0');

  int baselen = ext-origname;
  char *basename = (char *)malloc((baselen+1)*sizeof(char));
  strncpy(basename, origname, baselen);
  basename[baselen] = '\0';

  while ((fd == -1) && (lasterror == EEXIST)) {
    // Failed  because the file already existed. Try the next name.
    if (*newname)
      free(*newname);
    int namelen = strlen(basename) + strlen(ext) + 32;
    *newname = (char *)malloc(namelen*sizeof(char));
    if (!*newname)
      break;
    snprintf(*newname, namelen, "%s-%d%s", basename, i++, ext);

    RTMP_Log(RTMP_LOGDEBUG, "Trying next file: %s", *newname);
#ifdef WIN32
    fd = open(*newname, O_CREAT | O_EXCL | O_RDWR | O_BINARY);
#else
    fd = open(*newname, O_CREAT | O_EXCL | O_RDWR, 
              S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif
    lasterror = errno;
  }

  free(basename);

  if (fd == -1)
    return NULL;
  else
    return fdopen(fd, "r+b");
}

char *FormatYLEDateTime(char *datestring) {
  char *p = datestring;

  while (p) {
    p = strchr(p, 'T');
    if (p)
      *p = '-';
  }

  if (*datestring && 
      (datestring[strlen(datestring)-1] == 'Z' || 
       datestring[strlen(datestring)-1] == '-'))
    datestring[strlen(datestring)-1] = '\0';

  return datestring;
}

static size_t CollectStream(void *ptr, size_t size, size_t nmemb, void *stream) {
  MemoryBuffer *buf = stream;
  size_t bytes = size*nmemb;

  if (buf->len+bytes > buf->allocated) {
    if (bytes > 20*1024)
      buf->allocated += bytes;
    else
      buf->allocated += 20*1024;

    buf->ptr = realloc(buf->ptr, buf->allocated);
    if (!buf->ptr) {
      buf->allocated = 0;
      buf->len = 0;
      return 0;
    }
  }

  memcpy(buf->ptr+buf->len, ptr, bytes);
  buf->len += bytes;

  return bytes;
}

char *GetXMLNodeContent(const char *xmldoc, const char *node) {
  char *p, *pend, *ret = NULL;
  size_t contentlen ;

  if (!xmldoc)
    return NULL;

  p = strstr(xmldoc, node);
  if (p) {
    p += strlen(node);
    pend = strchr(p, '<');
    if (pend) {
      contentlen = pend-p;
      ret = malloc(contentlen+1);
      strncpy(ret, p, contentlen);
      ret[contentlen] = '\0';

      p = ret;
      while ((p = strstr(p, "&amp;"))) {
        size_t len = strlen(p+5);
        p = memmove(p+1, p+5, len);
        p[len] = '\0';
      }
    }
  }

  return ret;
}

void ConstructAkamaiParams(const char *streamname,
                           const char *authparams,
                           const char *appname,
                           const char *hostname,
                           const char *hostip,
                           AVal *playpath,
                           AVal *app,
                           AVal *tcUrl) {
  char *pp = malloc(strlen(streamname) + strlen(authparams) + 2);
  if (!pp) return;
  sprintf(pp, "%s?%s", streamname, authparams);
  STR2AVAL((*playpath), pp);

  char *appstr = malloc(strlen(appname) + strlen(hostname) + 
                        strlen(authparams) + 14);
  if (!appstr) return;
  sprintf(appstr, "%s?_fcs_vhost=%s&%s", appname, hostname, authparams);
  STR2AVAL((*app), appstr);

  char *tc = malloc(strlen(hostip) + strlen(appname) + 
                    strlen(hostname) + strlen(authparams) + 32);
  if (!tc) return;
  sprintf(tc, "rtmp://%s:1935/%s?_fcs_vhost=%s&%s", 
          hostip, appname, hostname, authparams);
  STR2AVAL((*tcUrl), tc);
}

int LoadAkamaiStreamParameters(RTMP *r, const char *xmlurl) {
  char date[64];
  struct HTTP_ctx http = {0};
  MemoryBuffer buf;
  AVal playpath = {NULL, 0};
  AVal tcUrl = {NULL, 0};
  AVal app = {NULL, 0};
  AVal host = {NULL, 0};
  AVal AVNull = {NULL, 0};
  char *streamname, *edgehost, *appname, *authparams, *hostip;

  if (!xmlurl)
    return FALSE;

  buf.len = 0;
  buf.allocated = 10240;
  buf.ptr = malloc(buf.allocated);
  if (!buf.ptr)
    return FALSE;

  date[0] = '\0';
  http.date = date;
  http.data = &buf;

  char *httpurl = strdup(xmlurl);
  
  if (strncmp(httpurl, "rtmp://", 7) == 0)
    strncpy(httpurl, "http", 4);

  RTMP_Log(RTMP_LOGDEBUG, "Downloading Akamai XML from %s", httpurl);

  HTTPResult httpres = HTTP_get(&http, httpurl, CollectStream);
  if (httpres != HTTPRES_OK) {
    if (httpres == HTTPRES_NOT_FOUND)
      RTMP_Log(RTMP_LOGERROR, "XML file %s not found", httpurl);
    else if (httpres == HTTPRES_LOST_CONNECTION)
      RTMP_Log(RTMP_LOGERROR, "Connection lost while downloading %s", httpurl);
    else
      RTMP_Log(RTMP_LOGERROR, "Failed to download %s (HTTP error %d)", httpurl, http.status);

    FREE_AND_NULL(buf.ptr);
    FREE_AND_NULL(httpurl);
    return FALSE;
  }

  CollectStream("", 1, 1, &buf);

  RTMP_Log(RTMP_LOGDEBUG, "Akamai XML:\n%s\n", buf.ptr);

  streamname = GetXMLNodeContent(buf.ptr, "<streamName>");
  edgehost = GetXMLNodeContent(buf.ptr, "<hostname>");
  appname = GetXMLNodeContent(buf.ptr, "<appName>");
  authparams = GetXMLNodeContent(buf.ptr, "<authParams>");

  FREE_AND_NULL(buf.ptr);

  if (!streamname || !edgehost || !appname || !authparams) {
    RTMP_Log(RTMP_LOGDEBUG, "Hostname, appName, streamName or authParams missing!");
    return FALSE;
  }

  buf.len = 0;
  buf.allocated = 10240;
  buf.ptr = malloc(buf.allocated);
  if (!buf.ptr)
    return FALSE;

  http.size = 0;
  http.status = -1;
  http.data = &buf;

  char *identurl = malloc(strlen(edgehost)+18);
  sprintf(identurl, "http://%s/fcs/ident", edgehost);

  RTMP_Log(RTMP_LOGDEBUG, "Downloading Akamai IDENT from %s", identurl);
  
  httpres = HTTP_get(&http, identurl, CollectStream);
  if (httpres != HTTPRES_OK) {
    if (httpres == HTTPRES_NOT_FOUND)
      RTMP_Log(RTMP_LOGERROR, "IDENT file %s not found", identurl);
    else if (httpres == HTTPRES_LOST_CONNECTION)
      RTMP_Log(RTMP_LOGERROR, "Connection lost while downloading %s", identurl);
    else
      RTMP_Log(RTMP_LOGERROR, "Failed to download %s (HTTP error %d)", identurl, http.status);

    FREE_AND_NULL(buf.ptr);
    FREE_AND_NULL(identurl);
    return FALSE;
  }

  CollectStream("", 1, 1, &buf);

  RTMP_Log(RTMP_LOGDEBUG, "Akamai IDENT:\n%s\n", buf.ptr);

  hostip = GetXMLNodeContent(buf.ptr, "<ip>");
  FREE_AND_NULL(buf.ptr);

  if (!hostip) {
    RTMP_Log(RTMP_LOGERROR, "IP missing from IDENT response!");
  }

  ConstructAkamaiParams(streamname, authparams, appname, edgehost, hostip,
                        &playpath, &app, &tcUrl);
  STR2AVAL(host, strdup(edgehost));

#ifdef CRYPTO
  uint32_t swfsize = r->Link.SWFSize;
#else
  uint32_t swfsize = 0;
#endif

  // FIXME: the old values are leaked
  RTMP_SetupStream(r, RTMP_PROTOCOL_RTMP, &host, 1935, &AVNull, &playpath,
                   &tcUrl, &(r->Link.swfUrl), &(r->Link.pageUrl), &app, &AVNull,
                   &AVNull, swfsize, &(r->Link.flashVer),
                   &AVNull, 0.0, 0, (r->Link.lFlags & RTMP_LF_LIVE) != 0,
                   r->Link.timeout);

  AMFObjectProperty prop = {{0, 0}};
  prop.p_type = AMF_BOOLEAN;
  prop.p_vu.p_number = 0;
  AMF_Reset(&r->Link.extras);
  AMF_AddProp(&r->Link.extras, &prop);

  // Don't use the YLE-specific invoke handler anymore
  r->invokeHook = NULL;

  FREE_AND_NULL(identurl);
  FREE_AND_NULL(httpurl);
  FREE_AND_NULL(streamname);
  //FREE_AND_NULL(edgehost);
  FREE_AND_NULL(appname);
  FREE_AND_NULL(authparams);
  FREE_AND_NULL(hostip);

  return TRUE;
}

int YLE_ConnectRedirected(RTMP *r, double dSeek) {
  if (!r->m_yleStream)
    return FALSE;

  if (r->m_yleStream->tvFeeRequired && !r->m_yleStream->tvFeeActivated) {
    RTMP_Log(RTMP_LOGERROR, "This stream requires YLE Passi");
    RTMP_Log(RTMP_LOGERROR, "RTMP_Login at https://login.yle.fi/login/index.php?language=1&url=areena.yle.fi");
    RTMP_Log(RTMP_LOGERROR, "and retry by giving the value of ChatYleFi cookie (--ylePassi)\n");
    return FALSE;

  } else if (r->m_yleStream->redirected) {
    RTMP_Close(r);

    // r->Link has been updated in ExecuteInvokedMethod to point to another server.
    RTMP_Log(RTMP_LOGDEBUG, "Redirected. Reconnecting...");

    r->m_yleStream->redirected = FALSE;
    return RTMP_Connect(r, NULL) && RTMP_ConnectStream(r, dSeek);
  }

  return FALSE;
}

static int DateHasTime(const char *datestring) {
  // A date time has format 2010-04-15T11:22:07.
  // A date without time has format 2010-04-15T.
  return strchr(datestring, ':') != NULL;
}
